﻿using System;
using System.Collections;
using System.Collections.Generic;
using MathNet.Numerics.Interpolation;
using SpectrumAnalysis.Calculation;
using SpectrumAnalysis.Interfaces;

namespace SpectrumAnalysis.Data
{
    /// <summary>
    /// Represents a set of chromaticity coordinates
    /// </summary>
    public class Chromaticity : IChromaticity, IEnumerable<ChromaPoint>
    {
        public readonly InterpolationType _interpolation;
        public readonly IObserver Observer;
        public readonly Xy WhitePoint;

        public Chromaticity(IObserver so, Xy whitepoint, InterpolationType interpolation)
        {
            Observer = so;
            WhitePoint = whitepoint;
            CalculateData(Observer, WhitePoint, 
                out Wavelengths, out XArray, out YArray,
                out XOffsetArray, out YOffsetArray, out RArray, out ThetaArray);
            Count = Wavelengths.Length;
            _interpolation = interpolation;
        }

        //public Chromaticity(StandardObservers observer, Xy whitepoint)
        //{
        //    Observer = new StandardObserver(observer);
        //    WhitePoint = whitepoint;
        //    CalculateData(Observer, WhitePoint,
        //        out Wavelengths, out XArray, out YArray, 
        //        out XOffsetArray, out YOffsetArray, out RArray, out ThetaArray);
        //    Count = Wavelengths.Length;
        //}

        private void CalculateData(IObserver observer, Xy whitepoint, out double[] wavelengths,
            out double[] xArray, out double[] yArray, out double[] xOffsetArray, out double[] yOffsetArray,
            out double[] rArray, out double[] thetaArray)
        {
            
            var points = CalculateCoordinates(observer);
            var pointArr = points.ToArray();
            var len = pointArr.Length;
            
            wavelengths = new double[len];
            
            xArray = new double[len];
            yArray = new double[len];
            xOffsetArray = new double[len];
            yOffsetArray = new double[len];
            rArray = new double[len];
            thetaArray = new double[len];
            xArray = new double[len];
            yArray = new double[len];
            
            for (var i = 0; i < len; ++i)
            {
                var pa = pointArr[i];
                var nm = pa.Wavelength;
                Wavelengths[i] = nm;
                var x = pa.X;
                var y = pa.Y;
                xArray[i] = x;
                yArray[i] = y;
                
                var xy = new Xy(x, y);
                
                var xyTrans = xy.Shift(whitepoint);
                xOffsetArray[i] = xyTrans.X;
                yOffsetArray[i] = xyTrans.Y;
                var polar = new PolarCoord(xyTrans);
                rArray[i] = polar.R;
                thetaArray[i] = polar.Theta;

                
            }

            
                _cubicSpline = CubicSpline.InterpolateNaturalSorted(ThetaArray, Wavelengths);
           
                _linearSpline = LinearSpline.InterpolateSorted(ThetaArray, Wavelengths);
            


        }

        public readonly double[] Wavelengths;
        public readonly double[] XArray;
        public readonly double[] YArray;

        public readonly double[] XOffsetArray;
        public readonly double[] YOffsetArray;
        public readonly double[] RArray;
        public readonly double[] ThetaArray;
        

        public readonly int Count;
        private CubicSpline _cubicSpline;
        private LinearSpline _linearSpline;
        
        

        private double GetInterpolatedWavelengthLinear(double theta)
        {
            

            // Create Interpolator for this segment
            
            var wavelength = _linearSpline.Interpolate(theta);
            return wavelength;
        }

        private double GetInterpolatedWavelengthCubic(double theta)
        {
            
            var wavelength = _cubicSpline.Interpolate(theta);
            return wavelength;
        }

       

        private static List<WavelenXyz> CalculateCoordinates(IObserver obs)
        {
            var points = new List<WavelenXyz>();
            for (var i = 0; i < obs.GetWavelengths().Length; ++i)
            {
                var nm = obs.GetWavelengths()[i];
                var x = obs.GetXData()[i];
                var y = obs.GetYData()[i];
                var z = obs.GetZData()[i];
                var xyz = new Xyz(x, y, z);
                var xy = xyz.GetChromaticityCoordinate();
                var wlxy = new WavelenXyz(nm, xy.X, xy.Y, 0.0);
                points.Add(wlxy);
            }
            return points;
        }

        public ChromaPoint GetChromaPoint(int index)
        {
            var cp = new ChromaPoint(Wavelengths[index], XArray[index], YArray[index], XOffsetArray[index],
                YOffsetArray[index], RArray[index], ThetaArray[index]);
            return cp;
        }


        public IEnumerator<ChromaPoint> GetEnumerator()
        {
            var list = new List<ChromaPoint>();
            for (var i = 0; i < Count; ++i)
            {
                list.Add(new ChromaPoint(Wavelengths[i], XArray[i], YArray[i], XOffsetArray[i],
                    YOffsetArray[i], RArray[i], ThetaArray[i]));
            }
            return list.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        public IObserver GetObserver()
        {
            return Observer;
        }

        public Xy GetWhitePoint()
        {
            return WhitePoint;
        }

        public double[] GetWavelengths()
        {
            return Wavelengths;
        }

        public double[] GetXData()
        {
            return XArray;
        }

        public double[] GetYData()
        {
            return YArray;
        }

        public double[] GetThetaArray()
        {
            return ThetaArray;
        }

        public double GetCount() {
            return Wavelengths.Length;
        }

        public double GetInterpolatedWavelength(double theta)
        {
            switch (_interpolation)
            {
                case InterpolationType.LINEAR:
                    return GetInterpolatedWavelengthLinear(theta);
                case InterpolationType.CUBIC:
                    return GetInterpolatedWavelengthCubic(theta);
                case InterpolationType.COSINE:
                    return CosineInterpolate(ThetaArray, Wavelengths, theta);
                default:
                    throw new Exception("Unknown interpolation type: " + _interpolation);
            }
        }

        /// <summary>
        /// http://paulbourke.net/miscellaneous/interpolation/
        /// </summary>
        /// <param name="y1"></param>
        /// <param name="y2"></param>
        /// <param name="mu"></param>
        /// <returns></returns>
        protected double CosineInterpolate(
           double y1, double y2,
           double mu)
        {
            

            double mu2 = (1 - Math.Cos(mu * Math.PI)) / 2;
            return (y1 * (1 - mu2) + y2 * mu2);
        }

        protected double CosineInterpolate(double[] angleArr, double[] wlArr, double theta)
        {
            double x1 = 0;
            double x2 = 0;
            double y1 = 0;
            double y2 = 0;
            for (var i = 0; i < angleArr.Length - 1; ++i)
            {
                var angle = angleArr[i];
                var nextAngle = angleArr[i + 1];
                if (theta == angle)
                {
                    return angle;
                }
                if ((theta >= angle) && (theta < nextAngle))
                {
                    y1 = wlArr[i];
                    y2 = wlArr[i + 1];
                    x1 = angle;
                    x2 = nextAngle;
                    break;
                }
            }
            double delta = x2 - x1;
            double deltaMu = theta - x1;
            double mu = deltaMu / delta;
            return CosineInterpolate(y1, y2, mu);
        }
    }
}
