﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using Microsoft.VisualBasic.FileIO;

namespace SpectrumAnalysis.Data 
{
    public class Spectrum : IEnumerable<SpectrumPoint>
    {
       

        private readonly List<SpectrumPoint> _spectrumList = new List<SpectrumPoint>();

        public Spectrum() { }

        public Spectrum(String fileName, FileType fileType, SpectrumType spectrumType)
        {
            LoadFromFile(fileName, fileType, spectrumType);
        }

        public Spectrum(Stream stream, FileType fileType, SpectrumType spectrumType)
        {
            LoadFromStream(stream, fileType, spectrumType);
        }

        public List<SpectrumPoint> List
        {
            get
            {
                return _spectrumList;
            }

        }

        public List<Xy> ListXy()
        {
            var xyList = new List<Xy>();
            foreach (var spectrumPoint in List)
            {
                xyList.Add(new Xy(spectrumPoint.Wavelength, spectrumPoint.Transmittance));
            }
            return xyList;
        }

        public void LoadFromFile(String fileName, FileType fileType, SpectrumType spectrumType)
        {

            _spectrumList.Clear();
            using (var parser = new TextFieldParser(fileName))
            {
                DoParser(parser, fileType, spectrumType);
            }
        }

        public void LoadFromStream(Stream stream, FileType fileType, SpectrumType spectrumType)
        {
            _spectrumList.Clear();
            using (var parser = new TextFieldParser(stream))
            {
                DoParser(parser, fileType, spectrumType);
            }
        }

        private void DoParser(TextFieldParser parser, FileType fileType, SpectrumType spectrumType)
        {
            parser.TextFieldType = FieldType.Delimited;
            if (fileType == FileType.CSV)
            {
                parser.SetDelimiters(",");
            }
            else if (fileType == FileType.TEXT)
            {
                parser.SetDelimiters("\t");
            }
            else
            {
                throw new Exception("Unsupported file type: " + fileType);
            }
            parser.TrimWhiteSpace = true;
            parser.HasFieldsEnclosedInQuotes = false;
            parser.CommentTokens = new[] { "#" };
            while (!parser.EndOfData)
            {
                var values = parser.ReadFields();
                if (values != null)
                {
                    var wavelength = Double.Parse(values[0]);
                    double transmittance;
                    var yval = Double.Parse(values[1]);
                    if (spectrumType == SpectrumType.TRANSMITTANCE)
                    {
                        transmittance = yval / 100.0;
                    }
                    else
                    {
                        transmittance = Math.Pow(10, -yval);
                    }
                    _spectrumList.Add(new SpectrumPoint(wavelength, transmittance));
                }
            }
        }

        IEnumerator<SpectrumPoint> IEnumerable<SpectrumPoint>.GetEnumerator()
        {
            return _spectrumList.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return _spectrumList.GetEnumerator();
        }

        public override string ToString()
        {
            var sb = new StringBuilder();
            for (var i = 0; i < _spectrumList.Count; ++i)
            {
                var point = _spectrumList[i];
                sb.AppendFormat("{0} {1}\r\n", i+1, point);
            }
            return sb.ToString();
        }

        public int GetFirstRoundedWavelength()
        {
            return (int)Math.Round(_spectrumList[0].Wavelength);
        }
    }
}
