﻿using MathNet.Numerics.Interpolation;

namespace SpectrumAnalysis.Data
{
    /// <summary>
    /// Source for data: http://www.cis.rit.edu/research/mcsl2/online/CIE/all_1nm_data.xls
    /// </summary>
    public class StandardIlluminant
    {
        private readonly StandardIlluminants _illuminant;

        public StandardIlluminants Illuminant
        {
            get { return _illuminant; }
        }

        public StandardIlluminant(StandardIlluminants illum)
        {
            _illuminant = illum;
        }

        public double GetIlluminationForWavelength(double wavelength)
        {
            return _illuminant == StandardIlluminants.D65 ? GetD65ForWavelength(wavelength) : 1.0;
        }


        //private static double[] _d65;

        //private static double[] D65
        //{
        //    get
        //    {
        //        if (_d65 != null)
        //        {
        //            return _d65;
        //        }
        //        var arr = D65Array();
        //        var numEntries = arr.GetLength(0);
        //        _d65 = new double[1000];
        //        _d65.Initialize();
        //        for (var i = 0; i < numEntries; ++i)
        //        {
        //            _d65[(int) arr[i, 0]] = arr[i, 1];
        //        }
        //        return _d65;
        //    }
        //}

        private LinearSpline _interpolator;

        private LinearSpline Interpolator
        {
            get
            {
                if (_interpolator == null)
                {
                    var arr = D65Array();
                    var size = arr.GetLength(0);
                    var wavelengths = new double[arr.Length];
                    var values = new double[arr.Length];
                    for (var i = 0; i < size; ++i)
                    {
                        wavelengths[i] = arr[i, 0];
                        values[i] = arr[i, 1];
                    }
                    
                    _interpolator = LinearSpline.Interpolate(wavelengths, values);
                }
                return _interpolator;
            }
        }

        private double GetD65ForWavelength(double wavelength)
        {
            return Interpolator.Interpolate(wavelength);
        }

        /// <summary>
        /// Ref: https://www.semrock.com/Data/Sites/1/semrockpdfs/whitepaper_howtocalculateluminositywavelengthandpurity.pdf
        /// </summary>
        public static Xy D65Whitepoint1931
        {
            get { return new Xy(0.3127, 0.3290); }
        }

        /// <summary>
        /// Ref: https://engineering.purdue.edu/~bouman/ece637/notes/pdf/ColorSpaces.pdf
        /// </summary>
        public static Xyz D65Whitepoint1931Xyz
        {
            get { return new Xyz(0.3127, 0.3290, 0.3583); }
        }

        public static Xy D65Whitepoint1964
        {
            get { return new Xy(0.31382, 0.33100); }
        }

        public static readonly Xyz D65Whitepoint1931Norm = new Xyz(95.04, 100.00, 108.88);
        public static readonly Xyz D65Whitepoint1964Norm = new Xyz(94.81, 100.00, 107.32);

        private static double[,] D65Array()
        {
            double[,] theArray =
            {

                {400, 82.7549},
                {401, 83.628},
                {402, 84.5011},
                {403, 85.3742},
                {404, 86.2473},
                {405, 87.1204},
                {406, 87.9936},
                {407, 88.8667},
                {408, 89.7398},
                {409, 90.6129},
                {410, 91.486},
                {411, 91.6806},
                {412, 91.8752},
                {413, 92.0697},
                {414, 92.2643},
                {415, 92.4589},
                {416, 92.6535},
                {417, 92.8481},
                {418, 93.0426},
                {419, 93.2372},
                {420, 93.4318},
                {421, 92.7568},
                {422, 92.0819},
                {423, 91.4069},
                {424, 90.732},
                {425, 90.057},
                {426, 89.3821},
                {427, 88.7071},
                {428, 88.0322},
                {429, 87.3572},
                {430, 86.6823},
                {431, 88.5006},
                {432, 90.3188},
                {433, 92.1371},
                {434, 93.9554},
                {435, 95.7736},
                {436, 97.5919},
                {437, 99.4102},
                {438, 101.228},
                {439, 103.047},
                {440, 104.865},
                {441, 106.079},
                {442, 107.294},
                {443, 108.508},
                {444, 109.722},
                {445, 110.936},
                {446, 112.151},
                {447, 113.365},
                {448, 114.579},
                {449, 115.794},
                {450, 117.008},
                {451, 117.088},
                {452, 117.169},
                {453, 117.249},
                {454, 117.33},
                {455, 117.41},
                {456, 117.49},
                {457, 117.571},
                {458, 117.651},
                {459, 117.732},
                {460, 117.812},
                {461, 117.517},
                {462, 117.222},
                {463, 116.927},
                {464, 116.632},
                {465, 116.336},
                {466, 116.041},
                {467, 115.746},
                {468, 115.451},
                {469, 115.156},
                {470, 114.861},
                {471, 114.967},
                {472, 115.073},
                {473, 115.18},
                {474, 115.286},
                {475, 115.392},
                {476, 115.498},
                {477, 115.604},
                {478, 115.711},
                {479, 115.817},
                {480, 115.923},
                {481, 115.212},
                {482, 114.501},
                {483, 113.789},
                {484, 113.078},
                {485, 112.367},
                {486, 111.656},
                {487, 110.945},
                {488, 110.233},
                {489, 109.522},
                {490, 108.811},
                {491, 108.865},
                {492, 108.92},
                {493, 108.974},
                {494, 109.028},
                {495, 109.082},
                {496, 109.137},
                {497, 109.191},
                {498, 109.245},
                {499, 109.3},
                {500, 109.354},
                {501, 109.199},
                {502, 109.044},
                {503, 108.888},
                {504, 108.733},
                {505, 108.578},
                {506, 108.423},
                {507, 108.268},
                {508, 108.112},
                {509, 107.957},
                {510, 107.802},
                {511, 107.501},
                {512, 107.2},
                {513, 106.898},
                {514, 106.597},
                {515, 106.296},
                {516, 105.995},
                {517, 105.694},
                {518, 105.392},
                {519, 105.091},
                {520, 104.79},
                {521, 105.08},
                {522, 105.37},
                {523, 105.66},
                {524, 105.95},
                {525, 106.239},
                {526, 106.529},
                {527, 106.819},
                {528, 107.109},
                {529, 107.399},
                {530, 107.689},
                {531, 107.361},
                {532, 107.032},
                {533, 106.704},
                {534, 106.375},
                {535, 106.047},
                {536, 105.719},
                {537, 105.39},
                {538, 105.062},
                {539, 104.733},
                {540, 104.405},
                {541, 104.369},
                {542, 104.333},
                {543, 104.297},
                {544, 104.261},
                {545, 104.225},
                {546, 104.19},
                {547, 104.154},
                {548, 104.118},
                {549, 104.082},
                {550, 104.046},
                {551, 103.641},
                {552, 103.237},
                {553, 102.832},
                {554, 102.428},
                {555, 102.023},
                {556, 101.618},
                {557, 101.214},
                {558, 100.809},
                {559, 100.405},
                {560, 100},
                {561, 99.6334},
                {562, 99.2668},
                {563, 98.9003},
                {564, 98.5337},
                {565, 98.1671},
                {566, 97.8005},
                {567, 97.4339},
                {568, 97.0674},
                {569, 96.7008},
                {570, 96.3342},
                {571, 96.2796},
                {572, 96.225},
                {573, 96.1703},
                {574, 96.1157},
                {575, 96.0611},
                {576, 96.0065},
                {577, 95.9519},
                {578, 95.8972},
                {579, 95.8426},
                {580, 95.788},
                {581, 95.0778},
                {582, 94.3675},
                {583, 93.6573},
                {584, 92.947},
                {585, 92.2368},
                {586, 91.5266},
                {587, 90.8163},
                {588, 90.1061},
                {589, 89.3958},
                {590, 88.6856},
                {591, 88.8177},
                {592, 88.9497},
                {593, 89.0818},
                {594, 89.2138},
                {595, 89.3459},
                {596, 89.478},
                {597, 89.61},
                {598, 89.7421},
                {599, 89.8741},
                {600, 90.0062},
                {601, 89.9655},
                {602, 89.9248},
                {603, 89.8841},
                {604, 89.8434},
                {605, 89.8026},
                {606, 89.7619},
                {607, 89.7212},
                {608, 89.6805},
                {609, 89.6398},
                {610, 89.5991},
                {611, 89.4091},
                {612, 89.219},
                {613, 89.029},
                {614, 88.8389},
                {615, 88.6489},
                {616, 88.4589},
                {617, 88.2688},
                {618, 88.0788},
                {619, 87.8887},
                {620, 87.6987},
                {621, 87.2577},
                {622, 86.8167},
                {623, 86.3757},
                {624, 85.9347},
                {625, 85.4936},
                {626, 85.0526},
                {627, 84.6116},
                {628, 84.1706},
                {629, 83.7296},
                {630, 83.2886},
                {631, 83.3297},
                {632, 83.3707},
                {633, 83.4118},
                {634, 83.4528},
                {635, 83.4939},
                {636, 83.535},
                {637, 83.576},
                {638, 83.6171},
                {639, 83.6581},
                {640, 83.6992},
                {641, 83.332},
                {642, 82.9647},
                {643, 82.5975},
                {644, 82.2302},
                {645, 81.863},
                {646, 81.4958},
                {647, 81.1285},
                {648, 80.7613},
                {649, 80.394},
                {650, 80.0268},
                {651, 80.0456},
                {652, 80.0644},
                {653, 80.0831},
                {654, 80.1019},
                {655, 80.1207},
                {656, 80.1395},
                {657, 80.1583},
                {658, 80.177},
                {659, 80.1958},
                {660, 80.2146},
                {661, 80.4209},
                {662, 80.6272},
                {663, 80.8336},
                {664, 81.0399},
                {665, 81.2462},
                {666, 81.4525},
                {667, 81.6588},
                {668, 81.8652},
                {669, 82.0715},
                {670, 82.2778},
                {671, 81.8784},
                {672, 81.4791},
                {673, 81.0797},
                {674, 80.6804},
                {675, 80.281},
                {676, 79.8816},
                {677, 79.4823},
                {678, 79.0829},
                {679, 78.6836},
                {680, 78.2842},
                {681, 77.4279},
                {682, 76.5716},
                {683, 75.7153},
                {684, 74.859},
                {685, 74.0027},
                {686, 73.1465},
                {687, 72.2902},
                {688, 71.4339},
                {689, 70.5776},
                {690, 69.7213},
                {691, 69.9101},
                {692, 70.0989},
                {693, 70.2876},
                {694, 70.4764},
                {695, 70.6652},
                {696, 70.854},
                {697, 71.0428},
                {698, 71.2315},
                {699, 71.4203},
                {700, 71.6091},
                {701, 71.8831},
                {702, 72.1571},
                {703, 72.4311},
                {704, 72.7051},
                {705, 72.979},
                {706, 73.253},
                {707, 73.527},
                {708, 73.801},
                {709, 74.075},
                {710, 74.349},
                {711, 73.0745},
                {712, 71.8},
                {713, 70.5255},
                {714, 69.251},
                {715, 67.9765},
                {716, 66.702},
                {717, 65.4275},
                {718, 64.153},
                {719, 62.8785},
                {720, 61.604},
                {721, 62.4322},
                {722, 63.2603},
                {723, 64.0885},
                {724, 64.9166},
                {725, 65.7448},
                {726, 66.573},
                {727, 67.4011},
                {728, 68.2293},
                {729, 69.0574},
                {730, 69.8856},
                {731, 70.4057},
                {732, 70.9259},
                {733, 71.446},
                {734, 71.9662},
                {735, 72.4863},
                {736, 73.0064},
                {737, 73.5266},
                {738, 74.0467},
                {739, 74.5669},
                {740, 75.087},
                {741, 73.9376},
                {742, 72.7881},
                {743, 71.6387},
                {744, 70.4893},
                {745, 69.3398},
                {746, 68.1904},
                {747, 67.041},
                {748, 65.8916},
                {749, 64.7421},
                {750, 63.5927},
                {751, 61.8752},
                {752, 60.1578},
                {753, 58.4403},
                {754, 56.7229},
                {755, 55.0054},
                {756, 53.288},
                {757, 51.5705},
                {758, 49.8531},
                {759, 48.1356},
                {760, 46.4182},
                {761, 48.4569},
                {762, 50.4956},
                {763, 52.5344},
                {764, 54.5731},
                {765, 56.6118},
                {766, 58.6505},
                {767, 60.6892},
                {768, 62.728},
                {769, 64.7667},
                {770, 66.8054},
                {771, 66.4631},
                {772, 66.1209},
                {773, 65.7786},
                {774, 65.4364},
                {775, 65.0941},
                {776, 64.7518},
                {777, 64.4096},
                {778, 64.0673},
                {779, 63.7251},
                {780, 63.3828},
                {781, 63.4749},
                {782, 63.567},
                {783, 63.6592},
                {784, 63.7513},
                {785, 63.8434},
                {786, 63.9355},
                {787, 64.0276},
                {788, 64.1198},
                {789, 64.2119},
                {790, 64.304},
                {791, 63.8188},
                {792, 63.3336},
                {793, 62.8484},
                {794, 62.3632},
                {795, 61.8779},
                {796, 61.3927},
                {797, 60.9075},
                {798, 60.4223},
                {799, 59.9371},
                {800, 59.4519}

            };
            return theArray;
        }
    }
}
