﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using Microsoft.VisualBasic.FileIO;
using MathNet.Numerics.Interpolation;
using SpectrumAnalysis.Interfaces;
using SpectrumAnalysis.Utility;

namespace SpectrumAnalysis.Data
{
   

    /// <summary>
    /// Source for data: http://www.cis.rit.edu/research/mcsl2/online/CIE/all_1nm_data.xls
    /// </summary>
    public class StandardObserver : IObserver
    {
        public static StandardObserver CIE_1931_JUDD_VOS_1978
        {
            get { return new StandardObserver(StandardObservers.CIE_1931_JUDD_VOS_1978); }
        }

        public static StandardObserver CIE_1931
        {
            get { return new StandardObserver(StandardObservers.CIE_1931); }
        }

        public static StandardObserver CIE_1964
        {
            get { return new StandardObserver(StandardObservers.CIE_1964); }
        }

        public double[] Wavelengths;
        public double[] XData;
        public double[] YData;
        public double[] ZData;

        public readonly StandardObservers Observer;
        
        public StandardObserver(StandardObservers observer)
        {
            Observer = observer;
            var data = LoadObserverData(observer);
            var size = data.Length;
            Wavelengths = new double[size];
            XData = new double[size];
            YData = new double[size];
            ZData = new double[size];
            for (var i = 0; i < size; ++i )
            {
                var dat = data[i];
                Wavelengths[i] = dat.Wavelength;
                XData[i] = dat.X;
                YData[i] = dat.Y;
                ZData[i] = dat.Z;
            }
            //_xCubicInterpolator = GetCubicInterpolator(Wavelengths, XData);
            //_yCubicInterpolator = GetCubicInterpolator(Wavelengths, YData);
            //_zCubicInterpolator = GetCubicInterpolator(Wavelengths, ZData);
            //_xInterpolator = GetInterpolator(Wavelengths, XData);
            //_yInterpolator = GetInterpolator(Wavelengths, YData);
            //_zInterpolator = GetInterpolator(Wavelengths, ZData);
            _linearInterpolator = new TrilinearInterpolator(Wavelengths, XData, YData, ZData);
        }

        //private readonly WavelenXyz[] _observerData;
        //private readonly LinearSpline _xInterpolator;
        //private readonly LinearSpline _yInterpolator;
        //private readonly LinearSpline _zInterpolator;
        //private readonly CubicSpline _xCubicInterpolator;
        //private readonly CubicSpline _yCubicInterpolator;
        //private readonly CubicSpline _zCubicInterpolator;

        private readonly TrilinearInterpolator _linearInterpolator;
       

        private WavelenXyz[] LoadObserverData(StandardObservers observer)
        {
            var csv = "";
            switch (observer)
            {
                case StandardObservers.CIE_1931:
                    csv = Properties.Resources.CIE_1931;
                    break;
                case StandardObservers.CIE_1931_JUDD_VOS_1978:
                    csv = Properties.Resources.CIE_1931_JUDD_VOS_1978;
                    break;
                case StandardObservers.CIE_1964:
                    csv = Properties.Resources.CIE_1964;
                    break;
            }

            var data = new List<WavelenXyz>();
            using (var resourceStream = new StringReader(csv))
            {
                

                
                using (var parser = new TextFieldParser(resourceStream))
                {
                    parser.TextFieldType = FieldType.Delimited;
                    parser.SetDelimiters(",");
                    parser.TrimWhiteSpace = true;
                    parser.HasFieldsEnclosedInQuotes = false;
                    parser.CommentTokens = new[] {"#"};
                    while (!parser.EndOfData)
                    {
                        var values = parser.ReadFields();
                        if (values == null) continue;
                        var wavelength = Double.Parse(values[0]);
                        var x = Double.Parse(values[1]);
                        var y = Double.Parse(values[2]);
                        var z = Double.Parse(values[3]);
                        var wxyz = new WavelenXyz(wavelength, x, y, z);
                        data.Add(wxyz);
                    }
                }
               
            }
            return data.ToArray();
        }



        //private static LinearSpline GetInterpolator(IEnumerable<double> wavelengths, IEnumerable<double> values)
        //{
            
        //    return LinearSpline.Interpolate(wavelengths, values);
        //}

        //private static CubicSpline GetCubicInterpolator(IEnumerable<double> wavelengths,
        //    IEnumerable<double> values)
        //{
            

        //    return CubicSpline.InterpolateNatural(wavelengths, values);
        //}

        public StandardObservers GetObserver()
        {
            return Observer;
        }

        public double[] GetWavelengths()
        {
            return Wavelengths;
        }

        public double[] GetXData()
        {
            return XData;
        }

        public double[] GetYData()
        {
            return YData;
        }

        public double[] GetZData()
        {
            return ZData;
        }

        /// <summary>
        /// Uses linear interpolation
        /// </summary>
        /// <param name="nanometers"></param>
        /// <returns></returns>
        public Xyz GetXyzForWavelength(double nanometers)
        {
            //var x = _xInterpolator.Interpolate(nanometers);
            //var y = _yInterpolator.Interpolate(nanometers);
            //var z = _zInterpolator.Interpolate(nanometers);
            Vector<double> interpolatedVector = _linearInterpolator.Interpolate(nanometers);
            var xyz = new Xyz(interpolatedVector[0], interpolatedVector[1], interpolatedVector[2]);
            return xyz;
        }

        //public Xyz GetXyzForWavelengthCubic(double nanometers)
        //{
        //    var x = _xCubicInterpolator.Interpolate(nanometers);
        //    var y = _yCubicInterpolator.Interpolate(nanometers);
        //    var z = _zCubicInterpolator.Interpolate(nanometers);
        //    return new Xyz(x, y, z);
        //}

    }
}
