﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows;
using System.Windows.Documents;
using System.Windows.Media;
using Microsoft.Win32;
using SpectrumAnalysis.Calculation;
using SpectrumAnalysis.Data;
using SpectrumAnalysis.Plot;
using SpectrumAnalysis.Utility;
using SpectrumAnalysis.Wpf.Plot;
using SpectrumAnalysis.Wpf.Utility;

namespace SpectrumAnalysis.Wpf
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : INotifyPropertyChanged
    {
       

        public MainWindow()
        {
            InitializeComponent();
            Loaded += MainWindow_Loaded;
        }

        void MainWindow_Loaded(object sender, RoutedEventArgs e)
        {
            // Set the title to AssemblyTitle + version number
            Title = AppUtil.GetAppTitleAndVersion();
            MainGrid.DataContext = this;
            TransmittanceMenuItem.IsChecked = (AppSettings.LoadedSpectrumType == SpectrumType.TRANSMITTANCE);
            AbsorbanceMenuItem.IsChecked = (AppSettings.LoadedSpectrumType == SpectrumType.ABSORBANCE);

        }

        private Brush _myBrush;
        public Brush MyBrush
        {
            get { return _myBrush; }
            set
            {
                _myBrush = value;
                RaisePropertyChanged("MyBrush");
            }
        }

        public String Cie1931
        {
            get { return Path.Combine(AppSettings.ResourcesDirectory, "cie1931.png"); }
        }

        private string _shortFileName;
        public string ShortFileName
        {
            get { return _shortFileName; }
            set
            {
                _shortFileName = value;
                RaisePropertyChanged("ShortFileName");
            }
        }

        public string SaveFileName
        {
            get
            {
                if (ShortFileName == null)
                {
                    return "";
                }
                var dt = Now.ToString("s").Replace(":", ".");
                if (_isBatch)
                {
                    return "SpectrumAnalysis." + dt + ".txt";
                }
                return ShortFileName + "." + dt + ".output.txt";
            }
        }

        private Boolean _isBatch;
        private FileType _selectedFileType;
        

        private string LogInfo { get; set; }

        public StandardObservers Observer
        {
            get { return StandardObservers.CIE_1931; }
            
        }

        

        public StandardIlluminants MyIlluminant
        {
            get { return StandardIlluminants.E;}
            
        }

        

        

        private SpectrumViewModel _mySpectrumViewModel;
        public SpectrumViewModel MySpectrumViewModel
        {
            get { return _mySpectrumViewModel; }
            set
            {
                _mySpectrumViewModel = value;
                RaisePropertyChanged("MySpectrumViewModel");
            }
        }

        private CIE1931ViewModel _cie1931ViewModel;

        public CIE1931ViewModel Cie1931ViewModel
        {
            get { return _cie1931ViewModel; }
            set
            {
                _cie1931ViewModel = value;
                RaisePropertyChanged("Cie1931ViewModel");
            }
        }

       

        //private Spectrum _observedSpectrum;

        public double DominantWavelength { get; set; }
        public double ComplementaryWavelength { get; set; }
        public DateTime Now { get; set; }

        private Spectrum _spectrum;
        private Spectrum _resampledSpectrum;
        //private const StandardObservers Observer = StandardObservers.CIE_1964;

        private void LoadSpectrumCommon(String spectrumFileName, FileType fileType, SpectrumType spectrumType, 
            out StandardObserver observer,
            out Xy xy, out double domWl, out double compWl)
        {
             _spectrum = new Spectrum(spectrumFileName, fileType, spectrumType);
            var fileArr = spectrumFileName.Split('/', '\\');

            ShortFileName = fileArr[fileArr.Length - 1];

            // Calculate dominant wavelength
            var whitepoint = new WhitePoint(Observer, MyIlluminant);
            

            observer = new StandardObserver(Observer);

            var cubicCalc = CalculatorFactory.GetSpectrumCalculator(Algorithm.A, InterpolationType.CUBIC);
            
            // Try with resampled spectrum
            var resampler = new SpectrumResampler(_spectrum);

            

            _resampledSpectrum = resampler.Resample();
            var xyzResamp = cubicCalc.CalculateXyzFromSpectrum(_resampledSpectrum, Observer, MyIlluminant);
            var xyyResamp = xyzResamp.ConvertToXyY();
            var xyResamp = xyyResamp.GetXy();
            xy = xyResamp;
            string outCubicResamp;
            //DominantWavelength = cubicCalc.CalculateDominantWavelength(observer, whitepoint.Xy, xyResamp, out outCubicResamp);
            DomCompWavelength domComp = cubicCalc.CalcDomCompWavelength(observer, whitepoint.Xy, xyResamp, out outCubicResamp, true);
            DominantWavelength = domComp.DominantWavelength ?? 0;
            ComplementaryWavelength = domComp.ComplementaryWavelength ?? 0;

            domWl = DominantWavelength;
            compWl = ComplementaryWavelength;
        }

        private void LoadSpectrumFromFile(String spectrumFileName, FileType fileType)
        {
            
            StandardObserver observer;
            Xy xy;
            double domWl;
            double compWl;
            LogNoSave("=============================");
            var sb = new StringBuilder();
            Log(sb, true, false, String.Format("{0} {1}", AppUtil.GetAppTitle(),
               AppUtil.GetAppVersionLong()));
            Log(sb, true, false, AppUtil.GetCopyright());
            Now = DateTime.Now;

            Log(sb, true, true, Now.ToString("f"));
            LoadSpectrumCommon(spectrumFileName, fileType, AppSettings.LoadedSpectrumType, out observer, out xy, out domWl, out compWl);




            var whitepoint = new WhitePoint(Observer, MyIlluminant);
            SetPlots(observer, whitepoint, xy, domWl);





            


            
           
            Log(sb, "File name: {0}", ShortFileName);
            Log(sb, "File Type: {0}", fileType);
            Log(sb, "Spectrum Type: {0}", AppSettings.LoadedSpectrumType);
            Log(sb, "Illuminant: {0}", MyIlluminant);
            Log(sb, "Observer: {0}", Observer);
            Log(sb, "Spectrum resampled to 5 nm");
            Log(sb, "Cubic interpolation of xy");
            Log(sb, true, false, "White point: x = {0}, y = {1}", Util.Round(whitepoint.Xy.X, 6),
                Util.Round(whitepoint.Xy.Y, 6));
            Log(sb, true, false, "Color point: x = {0}, y = {1}", Util.Round(xy.X, 6), Util.Round(xy.Y, 6));
            
            Log(sb, true, false, "Dominant wavelength: {0} nm",
                Util.Round4(DominantWavelength));

            Log(sb, true, false, "Complementary wavelength: {0} nm", Util.Round4(ComplementaryWavelength));
            
            LogInfo = sb.ToString();

        }

        private void SetPlots(StandardObserver observer, WhitePoint whitepoint, Xy xy, double domWl)
        {
            var viewModel = new SpectrumViewModel(new PlotData(_spectrum.ListXy()),
                null,
                ShortFileName, DominantWavelength, observer,
                new StandardIlluminant(MyIlluminant));
            viewModel.ResampledSpectrum = _resampledSpectrum;
            SpectrumPlot.Model = viewModel.SpectrumPlotModel;


            var horseResampModel = new CIE1931ViewModel(null, null, ShortFileName, domWl, observer,
                new StandardIlluminant(MyIlluminant), xy, whitepoint.Xy);
            HorseshoePlot.Model = horseResampModel.CIE1931PlotModel;
        }

        private void BatchProcessFiles(String[] spectrumFileNames)
        {
            
            var sb = new StringBuilder();
            

            LogNoSave("=============================");
            Log(sb, true, false, String.Format("{0} {1}", AppUtil.GetAppTitle(),
                AppUtil.GetAppVersionLong()));
            Log(sb, true, false, AppUtil.GetCopyright());
            Now = DateTime.Now;

            Log(sb, true, true, Now.ToString("f"));
            Log(sb, "Spectrum Type: {0}", AppSettings.LoadedSpectrumType);
            Log(sb, "Illuminant: {0}", MyIlluminant);
            Log(sb, "Observer: {0}", Observer);
            Log(sb, "Spectrum resampled to 5 nm");
            Log(sb, "Cubic interpolation of xy");
            var whitepoint = new WhitePoint(Observer, MyIlluminant);
            Log(sb, false, false, "White point: x = {0}, y = {1}", Util.Round(whitepoint.Xy.X, 6),
                    Util.Round(whitepoint.Xy.Y, 6));
            StandardObserver observer = null;
            Xy xy = null;
            double domWl = 0.0;
            double compWl = 0.0;
            int numFiles = spectrumFileNames.Length;
            var csvData = new String[numFiles, 5];  // Filename, x, y, DominantWavelength, ComplementaryWavelength
            for (var i = 0; i < numFiles; ++i )
            {
                var fileName = spectrumFileNames[i];
                var ext = Path.GetExtension(fileName);
                if (ext != null && ext.ToLowerInvariant() == ".csv")
                {
                    _selectedFileType = FileType.CSV;
                }
                else if (ext != null && ext.ToLowerInvariant() == ".txt")
                {
                    _selectedFileType = FileType.TEXT;
                }


                Log(sb, false, false, "-----------------------------");
                LoadSpectrumCommon(fileName, _selectedFileType, AppSettings.LoadedSpectrumType, out observer,
                    out xy, out domWl, out compWl);

                
                Log(sb, true, false, "File name: {0}", ShortFileName);
                Log(sb, "File Type: {0}", _selectedFileType);


                Log(sb, true, false, "Color point: x = {0}, y = {1}", Util.Round(xy.X, 6), Util.Round(xy.Y, 6));

                Log(sb, true, false, "Dominant wavelength: {0} nm",
                    Util.Round4(DominantWavelength));

                Log(sb, true, false, "Complementary wavelength: {0} nm",
                    Util.Round4(ComplementaryWavelength));

                csvData[i, 0] = ShortFileName;
                CultureInfo ci = CultureInfo.InvariantCulture;
                csvData[i, 1] = xy.X.ToString("F6", ci);
                csvData[i, 2] = xy.Y.ToString("F6", ci);
                csvData[i, 3] = DominantWavelength.ToString("F4", ci);
                csvData[i, 4] = ComplementaryWavelength.ToString("F4", ci);

            }


            SetPlots(observer, whitepoint, xy, domWl);

            // Display a table of csv data
            Log(sb, false, false, "-----------------------------");
            Log(sb, false, false, "CSV Data:");
            Log(sb, "Filename,x,y,DominantWavelength,ComplementaryWavelength");
            for (var i = 0; i < numFiles; ++i)
            {
                Log(sb, "{0},{1},{2},{3},{4}", csvData[i, 0], csvData[i, 1], csvData[i, 2], csvData[i, 3], csvData[i, 4]);
            }
            LogInfo = sb.ToString();
        }


        private void Log(StringBuilder sb, bool bold, bool underline, string msg, params Object[] args)
        {
            if (args != null && args.Length > 0)
            {
                msg = String.Format(msg, args);
            }

            if (sb != null)
            {
                sb.AppendLine(msg);
            }

            var run = new Run(msg)
            {
                FontWeight = bold ? FontWeights.Bold : FontWeights.Normal
            };
            if (underline)
            {
                run.TextDecorations.Add(TextDecorations.Underline);
            }
            LogViewer.Blocks.Add(new Paragraph(run)
            {
                TextAlignment = TextAlignment.Left,
                Margin = new Thickness(0, 5, 0, 5)
                
            });
            LogScrollViewer.ScrollToBottom();
        }


        private void Log(StringBuilder sb, string formatMsg, params Object[] args)
        {
            Log(sb, false, false, formatMsg, args);
            
        }

        private void Log(StringBuilder sb, string msg)
        {
            Log(sb, false, false, msg, null);
        }

        private void Log(StringBuilder sb, bool bold, bool underline, string msg)
        {
            Log(sb, bold, underline, msg, null);
        }

        private void LogNoSave(string msg, params Object[] args)
        {
            var fmtMsg = (args != null) ? String.Format(msg, args) : msg;
            var run = new Run(fmtMsg)
            {
                FontWeight = FontWeights.Light,
                FontStyle = FontStyles.Italic
                
            };
           
            LogViewer.Blocks.Add(new Paragraph(run)
            {
                TextAlignment = TextAlignment.Left,
                Margin = new Thickness(0, 5, 0, 5)
            });
            LogScrollViewer.ScrollToBottom();
        }

        private void LogNoSave(string msg)
        {
            LogNoSave(msg, null);
        }

        private void ButtonCopyOutput_Click(object sender, RoutedEventArgs e)
        {
            try
            {
                var sb = new StringBuilder();
                foreach (var block in LogViewer.Blocks)
                {
                    if (block is Paragraph)
                    {
                        var para = (Paragraph)block;
                        foreach (var run in para.Inlines)
                        {
                            sb.Append(run.GetValue(Run.TextProperty));
                        }
                        sb.AppendLine();
                    }
                }
                Clipboard.SetText(sb.ToString());
            }
            catch (COMException)
            {
                // Avoid "Cliboard full" exception
            }
        }

        public event PropertyChangedEventHandler PropertyChanged;

        private void RaisePropertyChanged(string propertyName)
        {
            if (null != PropertyChanged)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        

       

        private void LoadSpectrumMenuItem_Click(object sender, RoutedEventArgs e)
        {
            //var loadSpectrum = new LoadSpectrum();
            //var dialogResult = loadSpectrum.ShowDialog();
            //if (dialogResult == true)
            //{
            //    _fullFileName = loadSpectrum.SpectrumFileName;
            //    _selectedFileType = loadSpectrum.SelectedFileType;
            //    _selectedSpectrumType = loadSpectrum.SelectedSpectrumType;
            //    LoadSpectrumFromFile(_fullFileName, _selectedFileType,
            //        _selectedSpectrumType);
            //}

            var dialog = new OpenFileDialog
            {
                Title = "Load Spectral Data",
                DefaultExt = ".csv",
                Filter = "Spectrum Files (CSV or TXT) (*.csv;*.txt)|*.csv;*.txt|CSV Files (*.csv)|*.csv|Text (Tab-separated) Files (*.txt)|*.txt|All Filles (*.*)|*.*",
                InitialDirectory = AppSettings.SpectrumLoadDirectory,
                RestoreDirectory = true,
                Multiselect = true
                
            };
            var result = dialog.ShowDialog();
            if (result == true)
            {
                String dir = null;
                if (dialog.FileNames.Length == 1)
                {
                    var fileName = dialog.FileNames[0];
                    var ext = Path.GetExtension(fileName);
                    if (ext != null && ext.ToLowerInvariant() == ".csv")
                    {
                        _selectedFileType = FileType.CSV;
                    }
                    else if (ext != null && ext.ToLowerInvariant() == ".txt")
                    {
                        _selectedFileType = FileType.TEXT;
                    }
                    
                    LoadSpectrumFromFile(dialog.FileNames[0], _selectedFileType);
                    _isBatch = false;
                    dir = Path.GetDirectoryName(dialog.FileNames[0]);
                    AppSettings.SpectrumLoadDirectory = dir;
                }
                else if (dialog.FileNames.Length > 1)
                {
                    
                    BatchProcessFiles(dialog.FileNames);
                    _isBatch = true;
                    dir = Path.GetDirectoryName(dialog.FileNames[0]);
                    AppSettings.SpectrumLoadDirectory = dir;
                }
            }
            
        }

       

        private void SaveCalculationResults_Click(object sender, RoutedEventArgs e)
        {
            if (String.IsNullOrWhiteSpace(LogInfo))
            {
                MessageBox.Show("There are no calculation results to save", "No results",
                    MessageBoxButton.OK, MessageBoxImage.Warning);
                return;
            }

            

            

            var dialog = new SaveFileDialog
            {
                Title = "Save Calculation Result",
                FileName = SaveFileName,
                DefaultExt = ".txt",
                OverwritePrompt = true,
                RestoreDirectory = true,
                InitialDirectory = AppSettings.InfoSaveDirectory,
                Filter = "Text Files (.txt)|*.txt"

            };

            var dialogResult = dialog.ShowDialog(this);
            if (dialogResult == true)
            {
                byte[] fileData = new UTF8Encoding(true).GetBytes(LogInfo);
                using (var saveFileStream = (FileStream) dialog.OpenFile())
                {
                    saveFileStream.Write(fileData, 0, fileData.Length);
                }
                LogNoSave("Saved file: {0}", dialog.FileName);
                // Set default directory to this
                AppSettings.InfoSaveDirectory = Path.GetDirectoryName(dialog.FileName);
                
            }
        }

        private void Quit_Click(object sender, RoutedEventArgs e)
        {
            Application.Current.Shutdown(0);
        }

        private void About_Click(object sender, RoutedEventArgs e)
        {
            var about = new About();
            about.ShowDialog();
        }

       

      

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            
            var viewModel = new SpectrumViewModel();

            SpectrumPlot.Model = viewModel.SpectrumPlotModel;

            var horseshoeModel = new CIE1931ViewModel();
            HorseshoePlot.Model = horseshoeModel.CIE1931PlotModel;
        }

        private void SaveResampledSpectrum_Click(object sender, RoutedEventArgs e)
        {
            if (CheckForNoSpectrum())
            {
                return;
            }

            var myDocuments = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            var fnArr = ShortFileName.Split('.');
            var fn = String.Join(".", fnArr, 0, fnArr.Length - 1) + "_resampled_5nm.csv";
            var dialog = new SaveFileDialog
            {
                Title = "Save Resampled Spectrum",
                FileName = fn,
                DefaultExt = ".csv",
                OverwritePrompt = true,
                RestoreDirectory = true,
                InitialDirectory = myDocuments,
                Filter = "CSV Files (.csv)|*.csv"
            };
            var dialogResult = dialog.ShowDialog(this);
            if (dialogResult == true)
            {
                var sb = new StringBuilder();
                foreach (var sp in _resampledSpectrum)
                {
                    sb.AppendFormat("{0},{1}\r\n", sp.Wavelength, sp.Transmittance * 100.0);
                }

                byte[] fileData = new UTF8Encoding(true).GetBytes(sb.ToString());
                using (var saveFileStream = (FileStream) dialog.OpenFile())
                {
                    saveFileStream.Write(fileData, 0, fileData.Length);
                }
                LogNoSave("Saved file: {0}", dialog.FileName);
            }
        }

        

     

        private void SetSpectrumType(bool transmittance)
        {
            AppSettings.LoadedSpectrumType = transmittance ? SpectrumType.TRANSMITTANCE : SpectrumType.ABSORBANCE;
            
        }

        private void TransmittanceMenuItem_Click(object sender, RoutedEventArgs e)
        {
            TransmittanceMenuItem.IsChecked = true;
            AbsorbanceMenuItem.IsChecked = false;
            SetSpectrumType(true);
            e.Handled = true;
        }

        private void AbsorbanceMenuItem_Click(object sender, RoutedEventArgs e)
        {
            TransmittanceMenuItem.IsChecked = false;
            AbsorbanceMenuItem.IsChecked = true;
            SetSpectrumType(false);
            e.Handled = true;
        }

        private void SaveResampledAbsorbanceSpectrum_Click(object sender, RoutedEventArgs e)
        {
            if (CheckForNoSpectrum())
            {
                return;
            }
            var myDocuments = AppSettings.GetMyDocumentsDirectory;
            var fnArr = ShortFileName.Split('.');
            var fn = String.Join(".", fnArr, 0, fnArr.Length - 1) + "_resampled_5nm_absorbance.csv";
            var dialog = new SaveFileDialog
            {
                Title = "Save Resampled Absorbance Spectrum",
                FileName = fn,
                DefaultExt = ".csv",
                OverwritePrompt = true,
                RestoreDirectory = true,
                InitialDirectory = myDocuments,
                Filter = "CSV Files (.csv)|*.csv"
            };
            var dialogResult = dialog.ShowDialog(this);
            if (dialogResult == true)
            {
                var sb = new StringBuilder();
                foreach (var sp in _resampledSpectrum)
                {
                    sb.AppendFormat("{0},{1}\r\n", sp.Wavelength, -Math.Log10(sp.Transmittance));
                }
                byte[] fileData = new UTF8Encoding(true).GetBytes(sb.ToString());
                using (var saveFileStream = (FileStream)dialog.OpenFile())
                {
                    saveFileStream.Write(fileData, 0, fileData.Length);
                }
                LogNoSave("Saved file: {0}", dialog.FileName);
            }

        }


        private bool CheckForNoSpectrum()
        {
            if (_spectrum == null)
            {
                MessageBox.Show("No spectrum is loaded", "No spectrum", MessageBoxButton.OK,
                    MessageBoxImage.Warning);
                return true;
            }
            return false;
        }

        private void MainGrid_Drop(object sender, DragEventArgs e)
        {
            e.Effects = DragDropEffects.None;
            var dataObj = e.Data as DataObject;
            if (dataObj != null)
            {
                var fileList = new List<string>();
                
                foreach (var droppedFile in dataObj.GetFileDropList())
                {
                    fileList.Add(droppedFile);
                }
                if (fileList.Count > 0)
                {
                    BatchProcessFiles(fileList.ToArray());
                }
            }
        }
    }
}
