﻿using System;
using System.Collections.Generic;
using OxyPlot;
using OxyPlot.Axes;
using OxyPlot.Series;
using SpectrumAnalysis.Data;
using SpectrumAnalysis.Plot;

namespace SpectrumAnalysis.Wpf.Plot
{
    public class CIE1931ViewModel
    {
        private readonly static DataPoint[] EmptyDataPoints = { new DataPoint(0, 0), new DataPoint(0.9, 0.9) };

        public CIE1931ViewModel()
        {

            var series = new LineSeries
            {
                ItemsSource = EmptyDataPoints,
                LineStyle = LineStyle.None,
                MarkerType = MarkerType.None

            };

            CIE1931PlotModel = new PlotModel
            {
                Title = "No Data"

            };
            CIE1931PlotModel.Series.Add(series);

            SetGrid(CIE1931PlotModel);

            // Force axes to uniform size
            CIE1931PlotModel.Axes[0].Maximum = 0.9;
            CIE1931PlotModel.Axes[1].Maximum = 0.9;
        }

        public CIE1931ViewModel(PlotData specDat, PlotData observedData, string fileName,
            double dominantWavelength, StandardObserver observer, StandardIlluminant illuminant, Xy colorPt, Xy whitePt)
        {

            CIE1931PlotModel = new PlotModel
            {
                Title = String.Format("{0}, Observer: {1}, Illuminant: {2}",
                    fileName, observer.Observer, illuminant.Illuminant),
            };

            // Offset from whitePt to colorPt
            Xy diff = new Xy(colorPt.X - whitePt.X, colorPt.Y - whitePt.Y);

            // 2d linear distance between them
            double diffMag = Math.Sqrt(diff.X * diff.X + diff.Y * diff.Y);

            // Normalized vector direction pointing from the whitepoint to through the  color point
            Xy norm = new Xy(diff.X / diffMag, diff.Y / diffMag);

            // Create line segment starting at whitePt along norm to intersection test against
            double testSegLength = 1.0;
            Xy seg = new Xy(whitePt.X + norm.X * testSegLength, whitePt.Y + norm.Y * testSegLength);


            // Generate line series for the CIE chromaticity locus
            //
            double precision = 100;
            double precisionStep = 1.0 / precision;

            var observerData = new List<DataPoint>();

            Xy lastPt = new Xy(0.0, 0.0);
            Xy locusPt = new Xy(0.0, 0.0);

            double domWL = 0.0;

            for (double wave = 360.0; wave <= 830.0; wave += precisionStep)
            {
                Xyz obs = observer.GetXyzForWavelength(wave);

                XyY p = obs.ConvertToXyY();

                Xy curPt = p.GetXy();

                observerData.Add(new DataPoint(p.X, p.Y));

                double locusX;
                double locusY;

                // Intersect [lastPoint, point] with [segStart, segEnd]
                if (LineSegmentIntersection(lastPt.X, lastPt.Y, curPt.X, curPt.Y,
                    whitePt.X, whitePt.Y, seg.X, seg.Y, out locusX, out locusY))
                {
                    locusPt = new Xy(locusX, locusY);
                    domWL = wave;
                }

                lastPt = curPt;
            }
            _dominantWavelength = dominantWavelength;


            var segmentData = new List<DataPoint>();
            segmentData.Add(new DataPoint(whitePt.X, whitePt.Y));
            segmentData.Add(new DataPoint(locusPt.X, locusPt.Y));

            var whitePtToLocus = new LineSeries
            {
                ItemsSource = segmentData,
                Color = OxyColors.Black,
                BrokenLineThickness = 1,

                MarkerType = MarkerType.None,
                LineStyle = LineStyle.Dash

            };

            CIE1931PlotModel.Series.Add(whitePtToLocus);

            var observerSeries = new LineSeries
            {
                ItemsSource = observerData,
                Color = OxyColors.Black,
                BrokenLineThickness = 1,

                Title = "CIE 1931 Spectral Locus",

                MarkerType = MarkerType.None,
                LineStyle = LineStyle.Solid

            };

            CIE1931PlotModel.Series.Add(observerSeries);

            //// Vertical line for Dom WL
            //var vertBarSeries = new RectangleBarSeries();






            CIE1931PlotModel.IsLegendVisible = true;

            SetGrid(CIE1931PlotModel);

            var domWLAnnot = new OxyPlot.Annotations.ArrowAnnotation
            {
                EndPoint = new DataPoint(locusPt.X, locusPt.Y),
                ArrowDirection = new ScreenVector(-norm.X, norm.Y),
                Text = String.Format("{0:000.00}nm", new Decimal(DominantWavelength)),
                Color = OxyColors.Black,
                TextPosition = new DataPoint(locusPt.X - 0.1, locusPt.Y + 0.1),
                FontSize = 20,
                FontWeight = FontWeights.Bold
            };
            CIE1931PlotModel.Annotations.Add(domWLAnnot);

            var whitePtAnnot = new OxyPlot.Annotations.PointAnnotation
            {
                X = whitePt.X,
                Y = whitePt.Y,
                Text = "White Point",
                Fill = OxyColors.Black,
                Stroke = OxyColors.White
            };
            CIE1931PlotModel.Annotations.Add(whitePtAnnot);

            var colorPtAnnot = new OxyPlot.Annotations.PointAnnotation
            {
                X = colorPt.X,
                Y = colorPt.Y,
                Text = "Color Point",
                Fill = OxyColors.Black,
                Stroke = OxyColors.White
            };
            CIE1931PlotModel.Annotations.Add(colorPtAnnot);
        }

        private void SetGrid(PlotModel model)
        {
            var yAxis = new LinearAxis
            {
                Position = AxisPosition.Left,
                MajorGridlineStyle = LineStyle.Solid,
                MajorGridlineThickness = 1,
                MajorGridlineColor = OxyColors.LightGray,
                MinorGridlineStyle = LineStyle.Dot,
                MinorGridlineThickness = 1,
                MinorGridlineColor = OxyColors.LightGray,
                Key = "xAxis"
            };
            var xAxis = new LinearAxis
            {
                Position = AxisPosition.Bottom,
                MajorGridlineStyle = LineStyle.Solid,
                MajorGridlineThickness = 1,
                MajorGridlineColor = OxyColors.LightGray,
                MinorGridlineStyle = LineStyle.Dot,
                MinorGridlineThickness = 1,
                MinorGridlineColor = OxyColors.LightGray,
                Key = "xAxis"


            };
            model.Axes.Add(yAxis);
            model.Axes.Add(xAxis);


            model.TitleFont = "Segoe UI, Arial";
            model.TitleFontSize = 12;
            model.TitleFontWeight = FontWeights.Bold;



        }

        private double _dominantWavelength;

        public double DominantWavelength
        {
            get { return _dominantWavelength; }
        }

        /////////////////////////////////////////////////////////////////////////////
        // lineSegmentIntersection()
        //
        //  public domain function by Darel Rex Finley, 2006
        //
        //  Determines the intersection point of the line segment defined by points A and B
        //  with the line segment defined by points C and D.
        //
        //  Returns YES if the intersection point was found, and stores that point in X,Y.
        //  Returns NO if there is no determinable intersection point, in which case X,Y will
        //  be unmodified.
        //
        bool LineSegmentIntersection(
            double Ax, double Ay,
            double Bx, double By,
            double Cx, double Cy,
            double Dx, double Dy,
            out double X, out double Y)
        {

            double distAB, theCos, theSin, newX, ABpos;

            //  Fail if either line segment is zero-length.
            if (Ax == Bx && Ay == By
                || Cx == Dx && Cy == Dy)
            {
                X = 0.0;
                Y = 0.0;
                return false;
            }

            //  Fail if the segments share an end-point.
            if (Ax == Cx && Ay == Cy ||
                Bx == Cx && By == Cy ||
                Ax == Dx && Ay == Dy ||
                Bx == Dx && By == Dy)
            {
                X = 0.0;
                Y = 0.0;
                return false;
            }

            //  (1) Translate the system so that point A is on the origin.
            Bx -= Ax; By -= Ay;
            Cx -= Ax; Cy -= Ay;
            Dx -= Ax; Dy -= Ay;

            //  Discover the length of segment A-B.
            distAB = Math.Sqrt(Bx * Bx + By * By);

            //  (2) Rotate the system so that point B is on the positive X axis.
            theCos = Bx / distAB;
            theSin = By / distAB;

            newX = Cx * theCos + Cy * theSin;
            Cy = Cy * theCos - Cx * theSin; Cx = newX;

            newX = Dx * theCos + Dy * theSin;
            Dy = Dy * theCos - Dx * theSin; Dx = newX;

            //  Fail if segment C-D doesn't cross line A-B.
            if (Cy < 0.0 && Dy < 0.0 || Cy >= 0.0 && Dy >= 0.0)
            {
                X = 0.0;
                Y = 0.0;
                return false;
            }

            //  (3) Discover the position of the intersection point along line A-B.
            ABpos = Dx + (Cx - Dx) * Dy / (Dy - Cy);

            //  Fail if segment C-D crosses line A-B outside of segment A-B.
            if (ABpos < 0.0 || ABpos > distAB)
            {
                X = 0.0;
                Y = 0.0;
                return false;
            }

            //  (4) Apply the discovered position to line A-B in the original coordinate system.
            X = Ax + ABpos * theCos;
            Y = Ay + ABpos * theSin;

            //  Success.
            return true;
        }

        public PlotModel CIE1931PlotModel { get; private set; }
    }
}
