﻿using System;
using System.Collections.Generic;
using OxyPlot;
using OxyPlot.Axes;
using OxyPlot.Series;
using SpectrumAnalysis.Calculation;
using SpectrumAnalysis.Data;
using SpectrumAnalysis.Plot;

namespace SpectrumAnalysis.Wpf.Plot
{
    public class SpectrumViewModel
    {
        private readonly static DataPoint[] EmptyDataPoints = {new DataPoint(400, 0), new DataPoint(700, 1)};
       
        public SpectrumViewModel()
        {
           
            var series = new LineSeries
            {
                ItemsSource = EmptyDataPoints,
                 LineStyle = LineStyle.None,
                 MarkerType = MarkerType.None

            };
            SpectrumPlotModel = new PlotModel
            {
                Title = "No Data"
            };
            SpectrumPlotModel.Series.Add(series);

            SetGrid(SpectrumPlotModel);
            
        }
        public SpectrumViewModel(PlotData specDat, PlotData observedData, string fileName, 
            double dominantWavelength, StandardObserver observer, StandardIlluminant illuminant) 
        {
            // Remove spectral data outside the range 400-700
            var truncatedDataPoints = new List<Xy>();
            foreach (var dp in specDat.DataPoints)
            {
                if (dp.X >= SpectrumResampler.MinWavelength && dp.X <= SpectrumResampler.MaxWavelength)
                {
                    truncatedDataPoints.Add(new Xy(dp.X, dp.Y));
                }
            }
            var truncatedPlotData = new PlotData(truncatedDataPoints);

            OxyColor color = OxyColors.Green;
            if (dominantWavelength >= 578.6 && dominantWavelength <= 579.0)
            {
                color = OxyColors.Orange;
            }
            else if (dominantWavelength >= 579.0)
            {
                color = OxyColors.Red;
            }
            var spectrumSeries = new LineSeries
            {
                ItemsSource = truncatedPlotData.DataPoints,
                MarkerType = MarkerType.Circle,
                MarkerFill = color,
                MarkerStroke = color,
                MarkerStrokeThickness = 1,
                MarkerSize = 2,
                LineStyle = LineStyle.None,
                //BrokenLineThickness = 2,
                //BrokenLineStyle = LineStyle.Solid,
                Color = color,
                Title = "Transmittance Spectrum",
                Tag = "Spectrum"
            };
           
            
           

            SpectrumPlotModel = new PlotModel
            {
                Title = String.Format("{0}, Observer: {1}, Illuminant: {2}",
                fileName, observer.Observer, illuminant.Illuminant),
               
                
            };
            SpectrumPlotModel.Series.Add(spectrumSeries);
           
            


            if (observedData != null)
            {
                var observerData = new List<DataPoint>();
                var observerX = new List<DataPoint>();
                var observerY = new List<DataPoint>();
                var observerZ = new List<DataPoint>();

                foreach (var dp in truncatedPlotData.DataPoints)
                {
                    var wavelen = dp.X;
                    Xyz obs = observer.GetXyzForWavelength(wavelen);
                    var totalObs = (obs.X + obs.Y + obs.Z)/3.0;
                    observerData.Add(new DataPoint(wavelen, totalObs));
                    observerX.Add(new DataPoint(wavelen, obs.X));
                    observerY.Add(new DataPoint(wavelen, obs.Y));
                    observerZ.Add(new DataPoint(wavelen, obs.Z));
                }
                
                var observerIllumSeries = new LineSeries
                {
                    ItemsSource = observerData,
                    Title = "Total observer*illumn function",
                    
                    BrokenLineThickness = 1,
                    Color = OxyColors.Gray,
                    MarkerType = MarkerType.None,
                    LineStyle = LineStyle.Solid
                    
                };

                var xSeries = new LineSeries
                {
                    ItemsSource = observerX,
                    Title = "X",
                    
                    BrokenLineThickness = 1,
                    Color = OxyColors.Red,
                    MarkerType = MarkerType.None,
                    LineStyle = LineStyle.Solid
                    
                };

                var ySeries = new LineSeries
                {
                    ItemsSource = observerY,
                    Title = "Y",
                    
                    BrokenLineThickness = 1,
                    Color = OxyColors.Green,
                    MarkerType = MarkerType.None,
                    LineStyle = LineStyle.Solid
                    
                };

                var zSeries = new LineSeries
                {
                    ItemsSource = observerData,
                    Title = "Z",
                    
                    BrokenLineThickness = 1,
                    Color = OxyColors.Blue,
                    MarkerType = MarkerType.None,
                    LineStyle = LineStyle.Solid
                    
                };

                
                var observerSeries = new LineSeries
                {
                    ItemsSource = observedData.DataPoints,
                    Color = OxyColors.Purple,
                    BrokenLineThickness = 1,
                    
                    Title = "Observed Light Intensity",
                    MarkerType = MarkerType.None,
                    LineStyle = LineStyle.Solid
                    
                };


                SpectrumPlotModel.Series.Add(observerSeries);
                SpectrumPlotModel.Series.Add(observerIllumSeries);
                SpectrumPlotModel.Series.Add(xSeries);
                SpectrumPlotModel.Series.Add(ySeries);
                SpectrumPlotModel.Series.Add(zSeries);
            }

            //// Vertical line for Dom WL
            //var vertBarSeries = new RectangleBarSeries();
            
            
            
           
            

            SpectrumPlotModel.IsLegendVisible = true;

            SetGrid(SpectrumPlotModel);

            var annot = new OxyPlot.Annotations.ArrowAnnotation();

           
            

            SpectrumPlotModel.Annotations.Add(annot);


            


            
            


        }

        private void SetGrid(PlotModel model)
        {
            var xAxis = new LinearAxis
            {
                Position = AxisPosition.Bottom,
                MajorGridlineStyle = LineStyle.Solid,
                MajorGridlineThickness = 1,
                MajorGridlineColor = OxyColors.LightGray,
                MinorGridlineStyle = LineStyle.Dot,
                MinorGridlineThickness = 1,
                MinorGridlineColor = OxyColors.LightGray,
                Key = "xAxis",
                TitleFont = "Segoe UI",
                TitleFontSize = 12
                
            };
            model.Axes.Add(xAxis);

            model.TitleFont = "Segoe UI, Arial";
            model.TitleFontSize = 12;
            model.TitleFontWeight = FontWeights.Bold;
        }

        private Spectrum _resampledSpectrum;
        public Spectrum ResampledSpectrum
        {
            get { return _resampledSpectrum; }
            set
            {
                
                _resampledSpectrum = value;
                var plotData = new PlotData(_resampledSpectrum.ListXy());
                var resampledSpectrumSeries = new LineSeries
                {
                    ItemsSource = plotData.DataPoints,
                    MarkerType = MarkerType.Diamond,
                    MarkerFill = OxyColors.Violet, 
                    MarkerStroke = OxyColors.Violet,
                    MarkerStrokeThickness = 1,
                    MarkerSize = 2,
                    LineStyle = LineStyle.None,
                    Color = OxyColors.Violet,
                    Title = "Resampled Spectrum",
                    Tag = "ResampledSpectrum"
                };
                SpectrumPlotModel.Series.Add(resampledSpectrumSeries);
            }
        }

        public PlotModel SpectrumPlotModel { get; private set; }
    }
}
